@@ -27,6 +27,9 @@ import sys
 from threading import Thread, Event, RLock, Condition
 import time
 import ssl
+import weakref
+import random
+import itertools
 
 if 'gevent.monkey' in sys.modules:
     from gevent.queue import Queue, Empty
@@ -42,11 +45,15 @@ from cassandra.protocol import (ReadyMessage, AuthenticateMessage, OptionsMessag
                                 AuthResponseMessage, AuthChallengeMessage,
                                 AuthSuccessMessage, ProtocolException,
                                 RegisterMessage, ReviseRequestMessage)
+from cassandra.segment import SegmentCodec, CrcException
 from cassandra.util import OrderedDict
 from cassandra.shard_info import ShardingInfo
 
 log = logging.getLogger(__name__)
 
+segment_codec_no_compression = SegmentCodec()
+segment_codec_lz4 = None
+
 # We use an ordered dictionary and specifically add lz4 before
 # snappy so that lz4 will be preferred. Changing the order of this
 # will change the compression preferences for the driver.
@@ -88,6 +95,7 @@ else:
         return lz4_block.decompress(byts[3::-1] + byts[4:])
 
     locally_supported_compressions['lz4'] = (lz4_compress, lz4_decompress)
+    segment_codec_lz4 = SegmentCodec(lz4_compress, lz4_decompress)
 
 try:
     import snappy
@@ -109,6 +117,10 @@ HEADER_DIRECTION_FROM_CLIENT = 0x00
 HEADER_DIRECTION_TO_CLIENT = 0x80
 HEADER_DIRECTION_MASK = 0x80
 
+# shard aware default for opening per shard connection
+DEFAULT_LOCAL_PORT_LOW = 49152
+DEFAULT_LOCAL_PORT_HIGH = 65535
+
 frame_header_v1_v2 = struct.Struct('>BbBi')
 frame_header_v3 = struct.Struct('>BhBi')
 
@@ -426,6 +438,10 @@ class ProtocolError(Exception):
     pass
 
 
+class CrcMismatchException(ConnectionException):
+    pass
+
+
 class ContinuousPagingState(object):
     """
      A class for specifying continuous paging state, only supported starting with DSE_V2.
@@ -601,6 +617,71 @@ else:
     int_from_buf_item = ord
 
 
+class _ConnectionIOBuffer(object):
+    """
+    Abstraction class to ease the use of the different connection io buffers. With
+    protocol V5 and checksumming, the data is read, validated and copied to another
+    cql frame buffer.
+    """
+    _io_buffer = None
+    _cql_frame_buffer = None
+    _connection = None
+    _segment_consumed = False
+
+    def __init__(self, connection):
+        self._io_buffer = io.BytesIO()
+        self._connection = weakref.proxy(connection)
+
+    @property
+    def io_buffer(self):
+        return self._io_buffer
+
+    @property
+    def cql_frame_buffer(self):
+        return self._cql_frame_buffer if self.is_checksumming_enabled else \
+            self._io_buffer
+
+    def set_checksumming_buffer(self):
+        self.reset_io_buffer()
+        self._cql_frame_buffer = io.BytesIO()
+
+    @property
+    def is_checksumming_enabled(self):
+        return self._connection._is_checksumming_enabled
+
+    @property
+    def has_consumed_segment(self):
+        return self._segment_consumed;
+
+    def readable_io_bytes(self):
+        return self.io_buffer.tell()
+
+    def readable_cql_frame_bytes(self):
+        return self.cql_frame_buffer.tell()
+
+    def reset_io_buffer(self):
+        self._io_buffer = io.BytesIO(self._io_buffer.read())
+        self._io_buffer.seek(0, 2)  # 2 == SEEK_END
+
+    def reset_cql_frame_buffer(self):
+        if self.is_checksumming_enabled:
+            self._cql_frame_buffer = io.BytesIO(self._cql_frame_buffer.read())
+            self._cql_frame_buffer.seek(0, 2)  # 2 == SEEK_END
+        else:
+            self.reset_io_buffer()
+
+
+class ShardawarePortGenerator:
+    @classmethod
+    def generate(cls, shard_id, total_shards):
+        start = random.randrange(DEFAULT_LOCAL_PORT_LOW, DEFAULT_LOCAL_PORT_HIGH)
+        available_ports = itertools.chain(range(start, DEFAULT_LOCAL_PORT_HIGH), range(DEFAULT_LOCAL_PORT_LOW, start))
+
+        for port in available_ports:
+            if port % total_shards == shard_id:
+                yield port
+
+
 class Connection(object):
 
     CALLBACK_ERR_THREAD_THRESHOLD = 100
@@ -625,6 +706,7 @@ class Connection(object):
 
     # The current number of operations that are in flight. More precisely,
     # the number of request IDs that are currently in use.
+    # This includes orphaned requests.
     in_flight = 0
 
     # Max concurrent requests allowed per connection. This is set optimistically high, allowing
@@ -642,6 +724,20 @@ class Connection(object):
     # request_ids set
     highest_request_id = 0
 
+    # Tracks the request IDs which are no longer waited on (timed out), but
+    # cannot be reused yet because the node might still send a response
+    # on this stream
+    orphaned_request_ids = None
+
+    # Set to true if the orphaned stream ID count cross configured threshold
+    # and the connection will be replaced
+    orphaned_threshold_reached = False
+
+    # If the number of orphaned streams reaches this threshold, this connection
+    # will become marked and will be replaced with a new connection by the
+    # owning pool (currently, only HostConnection supports this)
+    orphaned_threshold = 3  * max_in_flight // 4
+
     is_defunct = False
     is_closed = False
     lock = None
@@ -656,7 +752,6 @@ class Connection(object):
 
     allow_beta_protocol_version = False
 
-    _iobuf = None
     _current_frame = None
 
     _socket = None
@@ -667,14 +762,23 @@ class Connection(object):
     _check_hostname = False
     _product_type = None
 
+    _owning_pool = None
+
     shard_id = 0
     sharding_info = None
 
+    _is_checksumming_enabled = False
+
+    @property
+    def _iobuf(self):
+        # backward compatibility, to avoid any change in the reactors
+        return self._io_buffer.io_buffer
+
     def __init__(self, host='127.0.0.1', port=9042, authenticator=None,
                  ssl_options=None, sockopts=None, compression=True,
                  cql_version=None, protocol_version=ProtocolVersion.MAX_SUPPORTED, is_control_connection=False,
                  user_type_map=None, connect_timeout=None, allow_beta_protocol_version=False, no_compact=False,
-                 ssl_context=None):
+                 ssl_context=None, owning_pool=None, shard_id=None, total_shards=None):
 
         # TODO next major rename host to endpoint and remove port kwarg.
         self.endpoint = host if isinstance(host, EndPoint) else DefaultEndPoint(host, port)
@@ -693,9 +797,11 @@ class Connection(object):
         self.no_compact = no_compact
         self._push_watchers = defaultdict(set)
         self._requests = {}
-        self._iobuf = io.BytesIO()
+        self._io_buffer = _ConnectionIOBuffer(self)
         self._continuous_paging_sessions = {}
         self._socket_writable = True
+        self.orphaned_request_ids = set()
+        self._owning_pool = owning_pool
 
         if ssl_options:
             self._check_hostname = bool(self.ssl_options.pop('check_hostname', False))
@@ -722,6 +828,9 @@ class Connection(object):
 
         self.lock = RLock()
         self.connected_event = Event()
+        self.shard_id = shard_id
+        self.total_shards = total_shards
+        self.original_endpoint = self.endpoint
 
     @property
     def host(self):
@@ -784,6 +893,15 @@ class Connection(object):
         self._socket = self.ssl_context.wrap_socket(self._socket, **ssl_options)
 
     def _initiate_connection(self, sockaddr):
+        if self.shard_id is not None:
+            for port in ShardawarePortGenerator.generate(self.shard_id, self.total_shards):
+                try:
+                    self._socket.bind(('', port))
+                    break
+                except Exception as ex:
+                    log.debug("port=%d couldn't bind cause: %s", port, str(ex))
+            log.debug(f'connection (%r) port=%d should be shard_id=%d', id(self), port, port % self.total_shards)
+
         self._socket.connect(sockaddr)
 
     def _match_hostname(self):
@@ -804,6 +922,7 @@ class Connection(object):
     def _connect_socket(self):
         sockerr = None
         addresses = self._get_socket_addresses()
+        port = None
         for (af, socktype, proto, _, sockaddr) in addresses:
             try:
                 self._socket = self._socket_impl.socket(af, socktype, proto)
@@ -816,6 +935,8 @@ class Connection(object):
                 self._socket.settimeout(self.connect_timeout)
                 self._initiate_connection(sockaddr)
                 self._socket.settimeout(None)
+                local_addr = self._socket.getsockname()
+                log.debug('Connection %s %s:%s -> %s:%s', id(self), local_addr[0], local_addr[1], sockaddr[0], sockaddr[1])
                 if self._check_hostname:
                     self._match_hostname()
                 sockerr = None
@@ -834,6 +955,16 @@ class Connection(object):
             for args in self.sockopts:
                 self._socket.setsockopt(*args)
 
+    def _enable_compression(self):
+        if self._compressor:
+            self.compressor = self._compressor
+
+    def _enable_checksumming(self):
+        self._io_buffer.set_checksumming_buffer()
+        self._is_checksumming_enabled = True
+        self._segment_codec = segment_codec_lz4 if self.compressor else segment_codec_no_compression
+        log.debug("Enabling protocol checksumming on connection (%s).", id(self))
+
     def close(self):
         raise NotImplementedError()
 
@@ -936,7 +1067,14 @@ class Connection(object):
         # queue the decoder function with the request
         # this allows us to inject custom functions per request to encode, decode messages
         self._requests[request_id] = (cb, decoder, result_metadata)
-        msg = encoder(msg, request_id, self.protocol_version, compressor=self.compressor, allow_beta_protocol_version=self.allow_beta_protocol_version)
+        msg = encoder(msg, request_id, self.protocol_version, compressor=self.compressor,
+                      allow_beta_protocol_version=self.allow_beta_protocol_version)
+
+        if self._is_checksumming_enabled:
+            buffer = io.BytesIO()
+            self._segment_codec.encode(buffer, msg)
+            msg = buffer.getvalue()
+
         self.push(msg)
         return len(msg)
 
@@ -1015,7 +1153,7 @@ class Connection(object):
 
     @defunct_on_error
     def _read_frame_header(self):
-        buf = self._iobuf.getvalue()
+        buf = self._io_buffer.cql_frame_buffer.getvalue()
         pos = len(buf)
         if pos:
             version = int_from_buf_item(buf[0]) & PROTOCOL_VERSION_MASK
@@ -1031,29 +1169,62 @@ class Connection(object):
                 self._current_frame = _Frame(version, flags, stream, op, header_size, body_len + header_size)
         return pos
 
-    def _reset_frame(self):
-        self._iobuf = io.BytesIO(self._iobuf.read())
-        self._iobuf.seek(0, 2)  # io.SEEK_END == 2 (constant not present in 2.6)
-        self._current_frame = None
+    @defunct_on_error
+    def _process_segment_buffer(self):
+        readable_bytes = self._io_buffer.readable_io_bytes()
+        if readable_bytes >= self._segment_codec.header_length_with_crc:
+            try:
+                self._io_buffer.io_buffer.seek(0)
+                segment_header = self._segment_codec.decode_header(self._io_buffer.io_buffer)
+
+                if readable_bytes >= segment_header.segment_length:
+                    segment = self._segment_codec.decode(self._iobuf, segment_header)
+                    self._io_buffer._segment_consumed = True
+                    self._io_buffer.cql_frame_buffer.write(segment.payload)
+                else:
+                    # not enough data to read the segment. reset the buffer pointer at the
+                    # beginning to not lose what we previously read (header).
+                    self._io_buffer._segment_consumed = False
+                    self._io_buffer.io_buffer.seek(0)
+            except CrcException as exc:
+                # re-raise an exception that inherits from ConnectionException
+                raise CrcMismatchException(str(exc), self.endpoint)
+        else:
+            self._io_buffer._segment_consumed = False
 
     def process_io_buffer(self):
         while True:
+            if self._is_checksumming_enabled and self._io_buffer.readable_io_bytes():
+                self._process_segment_buffer()
+                self._io_buffer.reset_io_buffer()
+
+            if self._is_checksumming_enabled and not self._io_buffer.has_consumed_segment:
+                # We couldn't read an entire segment from the io buffer, so return
+                # control to allow more bytes to be read off the wire
+                return
+
             if not self._current_frame:
                 pos = self._read_frame_header()
             else:
-                pos = self._iobuf.tell()
+                pos = self._io_buffer.readable_cql_frame_bytes()
 
             if not self._current_frame or pos < self._current_frame.end_pos:
+                if self._is_checksumming_enabled and self._io_buffer.readable_io_bytes():
+                    # We have a multi-segments message and we need to read more
+                    # data to complete the current cql frame
+                    continue
+
                 # we don't have a complete header yet or we
                 # already saw a header, but we don't have a
                 # complete message yet
                 return
             else:
                 frame = self._current_frame
-                self._iobuf.seek(frame.body_offset)
-                msg = self._iobuf.read(frame.end_pos - frame.body_offset)
+                self._io_buffer.cql_frame_buffer.seek(frame.body_offset)
+                msg = self._io_buffer.cql_frame_buffer.read(frame.end_pos - frame.body_offset)
                 self.process_msg(frame, msg)
-                self._reset_frame()
+                self._io_buffer.reset_cql_frame_buffer()
+                self._current_frame = None
 
     @defunct_on_error
     def process_msg(self, header, body):
@@ -1070,11 +1241,22 @@ class Connection(object):
                 decoder = paging_session.decoder
                 result_metadata = None
             else:
+                need_notify_of_release = False
+                with self.lock:
+                    if stream_id in self.orphaned_request_ids:
+                        self.in_flight -= 1
+                        self.orphaned_request_ids.remove(stream_id)
+                        need_notify_of_release = True
+                if need_notify_of_release and self._owning_pool:
+                    self._owning_pool.on_orphaned_stream_released()
+
                 try:
                     callback, decoder, result_metadata = self._requests.pop(stream_id)
                 # This can only happen if the stream_id was
                 # removed due to an OperationTimedOut
                 except KeyError:
+                    with self.lock:
+                        self.request_ids.append(stream_id)
                     return
 
         try:
@@ -1189,11 +1371,19 @@ class Connection(object):
                             compression_type = k
                             break
 
-                # set the decompressor here, but set the compressor only after
-                # a successful Ready message
-                self._compression_type = compression_type
-                self._compressor, self.decompressor = \
-                    locally_supported_compressions[compression_type]
+                # If snappy compression is selected with v5+checksumming, the connection
+                # will fail with OTO. Only lz4 is supported
+                if (compression_type == 'snappy' and
+                        ProtocolVersion.has_checksumming_support(self.protocol_version)):
+                    log.debug("Snappy compression is not supported with protocol version %s and "
+                              "checksumming. Consider installing lz4. Disabling compression.", self.protocol_version)
+                    compression_type = None
+                else:
+                    # set the decompressor here, but set the compressor only after
+                    # a successful Ready message
+                    self._compression_type = compression_type
+                    self._compressor, self.decompressor = \
+                        locally_supported_compressions[compression_type]
 
         self._send_startup_message(compression_type, no_compact=self.no_compact)
 
@@ -1214,6 +1404,7 @@ class Connection(object):
     def _handle_startup_response(self, startup_response, did_authenticate=False):
         if self.is_defunct:
             return
+
         if isinstance(startup_response, ReadyMessage):
             if self.authenticator:
                 log.warning("An authentication challenge was not sent, "
@@ -1222,8 +1413,11 @@ class Connection(object):
                             self.authenticator.__class__.__name__)
 
             log.debug("Got ReadyMessage on new connection (%s) from %s", id(self), self.endpoint)
-            if self._compressor:
-                self.compressor = self._compressor
+            self._enable_compression()
+
+            if ProtocolVersion.has_checksumming_support(self.protocol_version):
+                self._enable_checksumming()
+
             self.connected_event.set()
         elif isinstance(startup_response, AuthenticateMessage):
             log.debug("Got AuthenticateMessage on new connection (%s) from %s: %s",
@@ -1235,6 +1429,10 @@ class Connection(object):
                           "if DSE authentication is configured with transitional mode" % (self.host,))
                 raise AuthenticationFailed('Remote end requires authentication')
 
+            self._enable_compression()
+            if ProtocolVersion.has_checksumming_support(self.protocol_version):
+                self._enable_checksumming()
+
             if isinstance(self.authenticator, dict):
                 log.debug("Sending credentials-based auth response on %s", self)
                 cm = CredentialsMessage(creds=self.authenticator)
