@@ -15,12 +15,13 @@
 """
 Connection pooling and host management.
 """
-
+from concurrent.futures import Future
 from functools import total_ordering
 import logging
 import socket
 import time
 import random
+import copy
 from threading import Lock, RLock, Condition
 import weakref
 try:
@@ -384,16 +385,35 @@ class HostConnection(object):
     _lock = None
     _keyspace = None
 
+    # If the number of excess connections exceeds the number of shards times
+    # the number below, all excess connections will be closed.
+    max_excess_connections_per_shard_multiplier = 3
+
     def __init__(self, host, host_distance, session):
         self.host = host
         self.host_distance = host_distance
         self._session = weakref.proxy(session)
         self._lock = Lock()
         # this is used in conjunction with the connection streams. Not using the connection lock because the connection can be replaced in the lifetime of the pool.
-        self._stream_available_condition = Condition(self._lock)
+        self._stream_available_condition = Condition(Lock())
         self._is_replacing = False
         self._connecting = set()
         self._connections = {}
+        # A pool of additional connections which are not used but affect how Scylla
+        # assigns shards to them. Scylla tends to assign the shard which has
+        # the lowest number of connections. If connections are not distributed
+        # evenly at the moment, we might need to open several dummy connections
+        # to other shards before Scylla returns a connection to the shards we are
+        # interested in.
+        # After we get at least one connection for each shard, we can close
+        # the additional connections.
+        self._excess_connections = set()
+        # Contains connections which shouldn't be used anymore
+        # and are waiting until all requests time out or complete
+        # so that we can dispose of them.
+        self._trash = set()
+        self._shard_connections_futures = []
+        self.advanced_shardaware_block_until = 0
 
         if host_distance == HostDistance.IGNORED:
             log.debug("Not opening connection to ignored host %s", self.host)
@@ -403,21 +423,20 @@ class HostConnection(object):
             return
 
         log.debug("Initializing connection for host %s", self.host)
-        first_connection = session.cluster.connection_factory(self.host.endpoint)
+        first_connection = session.cluster.connection_factory(self.host.endpoint, owning_pool=self)
         log.debug("First connection created to %s for shard_id=%i", self.host, first_connection.shard_id)
         self._connections[first_connection.shard_id] = first_connection
         self._keyspace = session.keyspace
 
         if self._keyspace:
             first_connection.set_keyspace_blocking(self._keyspace)
-
-        if first_connection.sharding_info:
-            self.host.sharding_info = weakref.proxy(first_connection.sharding_info)
-            self._open_connections_for_all_shards()
+        if first_connection.sharding_info and not self._session.cluster.shard_aware_options.disable:
+            self.host.sharding_info = first_connection.sharding_info
+            self._open_connections_for_all_shards(first_connection.shard_id)
 
         log.debug("Finished initializing connection for host %s", self.host)
 
-    def borrow_connection(self, timeout, routing_key=None):
+    def _get_connection_for_routing_key(self, routing_key=None):
         if self.is_shutdown:
             raise ConnectionException(
                 "Pool for %s is shutdown" % (self.host,), self.host)
@@ -426,7 +445,7 @@ class HostConnection(object):
             raise NoConnectionsAvailable()
 
         shard_id = None
-        if self.host.sharding_info and routing_key:
+        if not self._session.cluster.shard_aware_options.disable and self.host.sharding_info and routing_key:
             t = self._session.cluster.metadata.token_map.token_class.from_key(routing_key)
             shard_id = self.host.sharding_info.shard_id_from_token(t.value)
 
@@ -442,6 +461,19 @@ class HostConnection(object):
                     self.host,
                     routing_key
                 )
+                if conn.orphaned_threshold_reached and shard_id not in self._connecting:
+                    # The connection has met its orphaned stream ID limit
+                    # and needs to be replaced. Start opening a connection
+                    # to the same shard and replace when it is opened.
+                    self._connecting.add(shard_id)
+                    self._session.submit(self._open_connection_to_missing_shard, shard_id)
+                    log.debug(
+                        "Connection to shard_id=%i reached orphaned stream limit, replacing on host %s (%s/%i)",
+                        shard_id,
+                        self.host,
+                        len(self._connections.keys()),
+                        self.host.sharding_info.shards_count
+                    )
             elif shard_id not in self._connecting:
                 # rate controlled optimistic attempt to connect to a missing shard
                 self._connecting.add(shard_id)
@@ -454,32 +486,47 @@ class HostConnection(object):
                     self.host.sharding_info.shards_count
                 )
 
-        # we couldn't find a shard aware connection, let's pick a random one
-        # from our pool
-        if not conn:
-            conn = self._connections.get(random.choice(list(self._connections.keys())))
+        if conn and not conn.is_closed:
+            return conn
+        active_connections = [conn for conn in list(self._connections.values()) if not conn.is_closed]
+        if active_connections:
+            return random.choice(active_connections)
+        return random.choice(list(self._connections.values()))
 
+    def borrow_connection(self, timeout, routing_key=None):
+        conn = self._get_connection_for_routing_key(routing_key)
         start = time.time()
         remaining = timeout
+        last_retry = False
         while True:
+            if conn.is_closed:
+                # The connection might have been closed in the meantime - if so, try again
+                conn = self._get_connection_for_routing_key(routing_key)
             with conn.lock:
-                if conn.in_flight < conn.max_request_id:
+                if (not conn.is_closed or last_retry) and conn.in_flight < conn.max_request_id:
+                    # On last retry we ignore connection status, since it is better to return closed connection than
+                    #  raise Exception
                     conn.in_flight += 1
                     return conn, conn.get_request_id()
             if timeout is not None:
                 remaining = timeout - time.time() + start
                 if remaining < 0:
-                    break
+                    # When timeout reached we try to get connection last time and break if it fails
+                    if last_retry:
+                        break
+                    last_retry = True
+                    continue
             with self._stream_available_condition:
                 self._stream_available_condition.wait(remaining)
 
         raise NoConnectionsAvailable("All request IDs are currently in use")
 
-    def return_connection(self, connection):
-        with connection.lock:
-            connection.in_flight -= 1
-        with self._stream_available_condition:
-            self._stream_available_condition.notify()
+    def return_connection(self, connection, stream_was_orphaned=False):
+        if not stream_was_orphaned:
+            with connection.lock:
+                connection.in_flight -= 1
+            with self._stream_available_condition:
+                self._stream_available_condition.notify()
 
         if connection.is_defunct or connection.is_closed:
             if connection.signaled_error and not self.shutdown_on_error:
@@ -501,37 +548,57 @@ class HostConnection(object):
                 self.shutdown()
             else:
                 connection.close()
-                del self._connections[connection.shard_id]
                 with self._lock:
+                    if self.is_shutdown:
+                        return
+                    self._connections.pop(connection.shard_id, None)
                     if self._is_replacing:
                         return
                     self._is_replacing = True
                     self._session.submit(self._replace, connection)
+        else:
+            if connection in self._trash:
+                with connection.lock:
+                    if connection.in_flight == len(connection.orphaned_request_ids):
+                        with self._lock:
+                            if connection in self._trash:
+                                self._trash.remove(connection)
+                        log.debug("Closing trashed connection (%s) to %s", id(connection), self.host)
+                        connection.close()
+                return
+
+    def on_orphaned_stream_released(self):
+        """
+        Called when a response for an orphaned stream (timed out on the client
+        side) was received.
+        """
+        with self._stream_available_condition:
+            self._stream_available_condition.notify()
 
     def _replace(self, connection):
         with self._lock:
             if self.is_shutdown:
                 return
 
-        log.debug("Replacing connection (%s) to %s", id(connection), self.host)
-        try:
-            if connection.shard_id in self._connections.keys():
-                del self._connections[connection.shard_id]
-            if self.host.sharding_info:
-                self._connecting.add(connection.shard_id)
-                self._open_connection_to_missing_shard(connection.shard_id)
+            log.debug("Replacing connection (%s) to %s", id(connection), self.host)
+            try:
+                if connection.shard_id in self._connections.keys():
+                    del self._connections[connection.shard_id]
+                if self.host.sharding_info and not self._session.cluster.shard_aware_options.disable:
+                    self._connecting.add(connection.shard_id)
+                    self._session.submit(self._open_connection_to_missing_shard, connection.shard_id)
+                else:
+                    connection = self._session.cluster.connection_factory(self.host.endpoint, owning_pool=self)
+                    if self._keyspace:
+                        connection.set_keyspace_blocking(self._keyspace)
+                    self._connections[connection.shard_id] = connection
+            except Exception:
+                log.warning("Failed reconnecting %s. Retrying." % (self.host.endpoint,))
+                self._session.submit(self._replace, connection)
             else:
-                connection = self._session.cluster.connection_factory(self.host.endpoint)
-                if self._keyspace:
-                    connection.set_keyspace_blocking(self._keyspace)
-                self._connections[connection.shard_id] = connection
-        except Exception:
-            log.warning("Failed reconnecting %s. Retrying." % (self.host.endpoint,))
-            self._session.submit(self._replace, connection)
-        else:
-            with self._lock:
                 self._is_replacing = False
-                self._stream_available_condition.notify()
+                with self._stream_available_condition:
+                    self._stream_available_condition.notify()
 
     def shutdown(self):
         log.debug("Shutting down connections to %s", self.host)
@@ -540,13 +607,63 @@ class HostConnection(object):
                 return
             else:
                 self.is_shutdown = True
-            self._stream_available_condition.notify_all()
+            with self._stream_available_condition:
+                self._stream_available_condition.notify_all()
 
-        if self._connections:
-            for c in self._connections.values():
-                log.debug("Closing connection (%s) to %s", id(c), self.host)
-                c.close()
-            self._connections = {}
+            for future in self._shard_connections_futures:
+                future.cancel()
+
+            connections_to_close = self._connections.copy()
+            self._connections.clear()
+
+        # connection.close can call pool.return_connection, which will
+        #  obtain self._lock via self._stream_available_condition.
+        # So, it never should be called within self._lock context
+        for connection in connections_to_close.values():
+            log.debug("Closing connection (%s) to %s", id(connection), self.host)
+            connection.close()
+
+        self._close_excess_connections()
+
+        trash_conns = None
+        with self._lock:
+            if self._trash:
+                trash_conns = self._trash
+                self._trash = set()
+
+        if trash_conns:
+            for conn in trash_conns:
+                conn.close()
+
+    def _close_excess_connections(self):
+        with self._lock:
+            if not self._excess_connections:
+                return
+            conns = self._excess_connections.copy()
+            self._excess_connections.clear()
+
+        for c in conns:
+            log.debug("Closing excess connection (%s) to %s", id(c), self.host)
+            c.close()
+
+    def disable_advanced_shard_aware(self, secs):
+        log.warning("disabling advanced_shard_aware for %i seconds, could be that this client is behind NAT?", secs)
+        self.advanced_shardaware_block_until = max(time.time() + secs, self.advanced_shardaware_block_until)
+
+    def _get_shard_aware_endpoint(self):
+        if (self.advanced_shardaware_block_until and self.advanced_shardaware_block_until < time.time()) or \
+           self._session.cluster.shard_aware_options.disable_shardaware_port:
+            return None
+
+        endpoint = None
+        if self._session.cluster.ssl_options and self.host.sharding_info.shard_aware_port_ssl:
+            endpoint = copy.copy(self.host.endpoint)
+            endpoint._port = self.host.sharding_info.shard_aware_port_ssl
+        elif self.host.sharding_info.shard_aware_port:
+            endpoint = copy.copy(self.host.endpoint)
+            endpoint._port = self.host.sharding_info.shard_aware_port
+
+        return endpoint
 
     def _open_connection_to_missing_shard(self, shard_id):
         """
@@ -560,33 +677,141 @@ class HostConnection(object):
         NOTE: This is an optimistic implementation since we cannot control
         which shard we want to connect to from the client side and depend on
         the round-robin of the system.clients shard_id attribution.
+
+        If we get a duplicate connection to some shard, we put it into the
+        excess connection pool. The more connections a particular shard has,
+        the smaller the chance that further connections will be assigned
+        to that shard.
         """
         with self._lock:
             if self.is_shutdown:
                 return
+        shard_aware_endpoint = self._get_shard_aware_endpoint()
+        log.debug("shard_aware_endpoint=%r", shard_aware_endpoint)
 
-        conn = self._session.cluster.connection_factory(self.host.endpoint)
-        if not self.is_shutdown and conn.shard_id not in self._connections.keys():
+        if shard_aware_endpoint:
+            conn = self._session.cluster.connection_factory(shard_aware_endpoint, owning_pool=self,
+                                                            shard_id=shard_id,
+                                                            total_shards=self.host.sharding_info.shards_count)
+            conn.original_endpoint = self.host.endpoint
+        else:
+            conn = self._session.cluster.connection_factory(self.host.endpoint, owning_pool=self)
+
+        log.debug("Received a connection %s for shard_id=%i on host %s", id(conn), conn.shard_id, self.host)
+        if self.is_shutdown:
+            log.debug("Pool for host %s is in shutdown, closing the new connection (%s)", self.host, id(conn))
+            conn.close()
+            return
+
+        if shard_aware_endpoint and shard_id != conn.shard_id:
+            # connection didn't land on expected shared
+            # assuming behind a NAT, disabling advanced shard aware for a while
+            self.disable_advanced_shard_aware(10 * 60)
+
+        old_conn = self._connections.get(conn.shard_id)
+        if old_conn is None or old_conn.orphaned_threshold_reached:
             log.debug(
-                "New connection created to shard_id=%i on host %s",
+                "New connection (%s) created to shard_id=%i on host %s",
+                id(conn),
                 conn.shard_id,
                 self.host
             )
-            self._connections[conn.shard_id] = conn
+            old_conn = None
+            with self._lock:
+                if self.is_shutdown:
+                    conn.close()
+                    return
+                if conn.shard_id in self._connections.keys():
+                    # Move the current connection to the trash and use the new one from now on
+                    old_conn = self._connections[conn.shard_id]
+                    log.debug(
+                        "Replacing overloaded connection (%s) with (%s) for shard %i for host %s",
+                        id(old_conn),
+                        id(conn),
+                        conn.shard_id,
+                        self.host
+                    )
+                self._connections[conn.shard_id] = conn
+            if old_conn is not None:
+                remaining = old_conn.in_flight - len(old_conn.orphaned_request_ids)
+                if remaining == 0:
+                    log.debug(
+                        "Immediately closing the old connection (%s) for shard %i on host %s",
+                        id(old_conn),
+                        old_conn.shard_id,
+                        self.host
+                    )
+                    old_conn.close()
+                else:
+                    log.debug(
+                        "Moving the connection (%s) for shard %i to trash on host %s, %i requests remaining",
+                        id(old_conn),
+                        old_conn.shard_id,
+                        self.host,
+                        remaining,
+                    )
+                    with self._lock:
+                        if self.is_shutdown:
+                            old_conn.close()
+                        else:
+                            self._trash.add(old_conn)
             if self._keyspace:
-                self._connections[conn.shard_id].set_keyspace_blocking(self._keyspace)
+                with self._lock:
+                    if self.is_shutdown:
+                        conn.close()
+                    old_conn = self._connections.get(conn.shard_id)
+                    if old_conn:
+                        old_conn.set_keyspace_blocking(self._keyspace)
+            num_missing_or_needing_replacement = self.num_missing_or_needing_replacement
             log.debug(
-                "Connected to %s/%i shards on host %s (%i missing)",
+                "Connected to %s/%i shards on host %s (%i missing or needs replacement)",
                 len(self._connections.keys()),
                 self.host.sharding_info.shards_count,
                 self.host,
-                self.host.sharding_info.shards_count - len(self._connections.keys())
+                num_missing_or_needing_replacement
+            )
+            if num_missing_or_needing_replacement == 0:
+                log.debug(
+                    "All shards of host %s have at least one connection, closing %i excess connections",
+                    self.host,
+                    len(self._excess_connections)
+                )
+                self._close_excess_connections()
+        elif self.host.sharding_info.shards_count == len(self._connections.keys()) and self.num_missing_or_needing_replacement == 0:
+            log.debug(
+                "All shards are already covered, closing newly opened excess connection %s for host %s",
+                id(self),
+                self.host
             )
-        else:
             conn.close()
+        else:
+            if len(self._excess_connections) >= self._excess_connection_limit:
+                log.debug(
+                    "After connection %s is created excess connection pool size limit (%i) reached for host %s, closing all %i of them",
+                    id(conn),
+                    self._excess_connection_limit,
+                    self.host,
+                    len(self._excess_connections)
+                )
+                self._close_excess_connections()
+
+            log.debug(
+                "Putting a connection %s to shard %i to the excess pool of host %s",
+                id(conn),
+                conn.shard_id,
+                self.host
+            )
+            close_connection = False
+            with self._lock:
+                if self.is_shutdown:
+                    close_connection = True
+                else:
+                    self._excess_connections.add(conn)
+            if close_connection:
+                conn.close()
         self._connecting.discard(shard_id)
 
-    def _open_connections_for_all_shards(self):
+    def _open_connections_for_all_shards(self, skip_shard_id=None):
         """
         Loop over all the shards and try to open a connection to each one.
         """
@@ -594,9 +819,13 @@ class HostConnection(object):
             if self.is_shutdown:
                 return
 
-        for shard_id in range(self.host.sharding_info.shards_count):
-            self._connecting.add(shard_id)
-            self._session.submit(self._open_connection_to_missing_shard, shard_id)
+            for shard_id in range(self.host.sharding_info.shards_count):
+                if skip_shard_id is not None and skip_shard_id == shard_id:
+                    continue
+                future = self._session.submit(self._open_connection_to_missing_shard, shard_id)
+                if isinstance(future, Future):
+                    self._connecting.add(shard_id)
+                    self._shard_connections_futures.append(future)
 
     def _set_keyspace_for_all_conns(self, keyspace, callback):
         """
@@ -623,20 +852,29 @@ class HostConnection(object):
                 callback(self, errors)
 
         self._keyspace = keyspace
-        for conn in self._connections.values():
+        for conn in list(self._connections.values()):
             conn.set_keyspace_async(keyspace, connection_finished_setting_keyspace)
 
     def get_connections(self):
-        c = self._connections
-        return list(self._connections.values()) if c else []
+        connections = self._connections
+        return list(connections.values()) if connections else []
 
     def get_state(self):
-        in_flights = [c.in_flight for c in self._connections.values()]
+        in_flights = [c.in_flight for c in list(self._connections.values())]
         return {'shutdown': self.is_shutdown, 'open_count': self.open_count, 'in_flights': in_flights}
 
+    @property
+    def num_missing_or_needing_replacement(self):
+        return self.host.sharding_info.shards_count \
+            - sum(1 for c in self._connections.values() if not c.orphaned_threshold_reached)
+
     @property
     def open_count(self):
-        return sum([1 if c and not (c.is_closed or c.is_defunct) else 0 for c in self._connections.values()])
+        return sum([1 if c and not (c.is_closed or c.is_defunct) else 0 for c in list(self._connections.values())])
+
+    @property
+    def _excess_connection_limit(self):
+        return self.host.sharding_info.shards_count * self.max_excess_connections_per_shard_multiplier
 
 
 _MAX_SIMULTANEOUS_CREATION = 1
@@ -667,7 +905,7 @@ class HostConnectionPool(object):
 
         log.debug("Initializing new connection pool for host %s", self.host)
         core_conns = session.cluster.get_core_connections_per_host(host_distance)
-        self._connections = [session.cluster.connection_factory(host.endpoint)
+        self._connections = [session.cluster.connection_factory(host.endpoint, owning_pool=self)
                              for i in range(core_conns)]
 
         self._keyspace = session.keyspace
@@ -771,7 +1009,7 @@ class HostConnectionPool(object):
 
         log.debug("Going to open new connection to host %s", self.host)
         try:
-            conn = self._session.cluster.connection_factory(self.host.endpoint)
+            conn = self._session.cluster.connection_factory(self.host.endpoint, owning_pool=self)
             if self._keyspace:
                 conn.set_keyspace_blocking(self._session.keyspace)
             self._next_trash_allowed_at = time.time() + _MIN_TRASH_INTERVAL
@@ -831,9 +1069,10 @@ class HostConnectionPool(object):
 
         raise NoConnectionsAvailable()
 
-    def return_connection(self, connection):
+    def return_connection(self, connection, stream_was_orphaned=False):
         with connection.lock:
-            connection.in_flight -= 1
+            if not stream_was_orphaned:
+                connection.in_flight -= 1
             in_flight = connection.in_flight
 
         if connection.is_defunct or connection.is_closed:
@@ -869,6 +1108,13 @@ class HostConnectionPool(object):
             else:
                 self._signal_available_conn()
 
+    def on_orphaned_stream_released(self):
+        """
+        Called when a response for an orphaned stream (timed out on the client
+        side) was received.
+        """
+        self._signal_available_conn()
+
     def _maybe_trash_connection(self, connection):
         core_conns = self._session.cluster.get_core_connections_per_host(self.host_distance)
         did_trash = False
@@ -933,11 +1179,16 @@ class HostConnectionPool(object):
                 self.is_shutdown = True
 
         self._signal_all_available_conn()
-        for conn in self._connections:
-            conn.close()
-            self.open_count -= 1
 
-        for conn in self._trash:
+        connections_to_close = []
+        with self._lock:
+            connections_to_close.extend(self._connections)
+            self.open_count -= len(self._connections)
+            self._connections.clear()
+            connections_to_close.extend(self._trash)
+            self._trash.clear()
+
+        for conn in connections_to_close:
             conn.close()
 
     def ensure_core_connections(self):
